from os.path import dirname, join, expanduser
from zrp.validate import *
from .preprocessing import *
from .geo_geocoder import *
from .acs_mapper import *
from .base import BaseZRP
from .utils import *
import pandas as pd
import numpy as np
import warnings
import glob
import json
import sys
import os
import re

import warnings
warnings.filterwarnings(action='ignore')


class ZRP_Prepare(BaseZRP):
    """
    Prepares data to generate race & ethnicity proxies

    Parameters
    ------------
    file_path: str, optional
        Path where to put artifacts and other files generated during intermediate steps.
    """
    
    def __init__(self, file_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.params_dict =  kwargs

        
    def fit(self, input_data):
        if self.census_tract:
            tract_lengths =  input_data[self.census_tract].str.len()
            tract_len  = most_common(tract_lengths)
            if not (input_data[self.census_tract].apply(lambda x: str(x).isalnum()).any()):
                raise ValueError("Cannot provide non-numeric Census Tract code, please remove non-numeric census tract records.")
            if tract_len != 11:
                raise ValueError("Improper Census Tract format provided. The tool requires the full state fips, county fips, and tract format. (ie '06037311600')")

        if self.block_group:
            bg_lengths =  input_data[self.block_group].str.len()
            bg_len  = most_common(bg_lengths)
            if bg_len != 12:  
                raise ValueError("Improper Census Block Group format provided. The tool requires the full state fips, county fips, tract, and block group format. (ie '060373116003')")

    
    def transform(self, input_data):
        """
        Transforms the data
        
        Parameters
        ----------
        input_data: pd.Dataframe
            Dataframe to be transformed
        """  
        curpath = dirname(__file__)
        # Load Data
        try:
            data = input_data.copy()
            print("Data is loaded")
        except AttributeError:
            data = load_file(self.file_path)
            print("Data file is loaded")
            
        data_path = join(curpath, f'../data/processed')
        lookup_tables_config = load_json(join(data_path, "lookup_tables_config.json"))

        geo_folder = os.path.join(data_path, "geo", lookup_tables_config['geo_year'])
        acs_folder = os.path.join(data_path, 'acs', lookup_tables_config['acs_year'], lookup_tables_config['acs_span'])

        if not ((os.path.isdir(geo_folder)) &
                (os.path.isdir(acs_folder ))
               ):
            raise AssertionError("Missing required support files please see the README for how to download the support files: https://github.com/zestai/zrp/blob/main/README.rst#install ")
        if not ((len(os.listdir(geo_folder)) > 0) &
                (len(os.listdir(acs_folder)) > 0)):
            raise AssertionError("Missing required support files please see the README for how to download the support files: https://github.com/zestai/zrp/blob/main/README.rst#install ") 
        gen_process = ProcessStrings(file_path=self.file_path, **self.params_dict)
        gen_process.fit(data)
        data = gen_process.transform(data)
        
        print("")

        print("[Start] Preparing geo data")

        inv_state_map = load_json(join(data_path, "inv_state_mapping.json"))
        data['zest_in_state_fips'] = data[self.state].replace(inv_state_map)
        print("")

        geocode = ZGeo(file_path=self.file_path, **self.params_dict)
        geocode_out = [] 
        geo_grps = data.groupby([self.state])
        geo_dict = {}
        for s, g in geo_grps:
            geo_dict[s] = g
        gdkys = list(geo_dict.keys())
        print("  The following states are included in the data:", gdkys)
              
        if not set(gdkys) <= set(list(inv_state_map.keys())):
            raise ValueError("Provided unrecognizable state codes. Please use standard 2-letter abbreviation to indicate states to geocode, ex:'CA' for Californina")

        geo_out = [] 
        for s in tqdm(gdkys):
            print("   ... on state:", str(s))
            geo = inv_state_map[s].zfill(2)
            output = geocode.transform(geo_dict[s], geo, processed = True, replicate = True, save_table = True)
            geocode_out.append(output)
        if len(geocode_out) > 0:
            geo_coded = pd.concat(geocode_out)

            # append data unable to enter geo mapping
            geo_coded_keys = list(geo_coded.ZEST_KEY_COL.values)
            data_not_geo_coded = data[~data.index.isin(geo_coded_keys)]
            geo_coded = pd.concat([geo_coded, data_not_geo_coded])
        else:
            geo_coded = data
            geo_coded['GEOID'] = None
            geo_coded['GEOID_BG'] = None
            geo_coded['GEOID_CT'] = None
            geo_coded['GEOID_ZIP'] = None
            geo_coded["ZEST_KEY_COL"] = geo_coded.index 
        # replace GEOIDs with user-defined values where avaliable
        if self.block_group is not None and self.census_tract is not None:
            geo_coded = geo_coded.drop([self.block_group, self.census_tract], axis = 1)
            geo_coded = geo_coded.merge(data[[self.block_group, self.census_tract]], right_index = True, left_index = True, how = 'left')
            geo_coded['GEOID_BG'] = np.where((geo_coded[self.block_group].isna()) | (geo_coded[self.block_group].str.contains("None") | (geo_coded[self.block_group] == ''))
                                             ,geo_coded['GEOID_BG']
                                             ,geo_coded[self.block_group])
            geo_coded['GEOID_CT'] = np.where((geo_coded[self.census_tract].isna()) | (geo_coded[self.census_tract].str.contains("None") | (geo_coded[self.census_tract] == ''))
                                             ,geo_coded['GEOID_CT']
                                             ,geo_coded[self.census_tract])
            geo_coded = geo_coded.drop([self.block_group, self.census_tract], axis = 1) 
        elif self.block_group is not None:
            geo_coded = geo_coded.drop(self.block_group, axis = 1)
            geo_coded = geo_coded.merge(data[self.block_group], right_index = True, left_index = True, how = 'left')
            geo_coded['GEOID_BG'] = np.where((geo_coded[self.block_group].isna()) | (geo_coded[self.block_group].str.contains("None") | (geo_coded[self.block_group] == ''))
                                             ,geo_coded['GEOID_BG']
                                             ,geo_coded[self.block_group])
            geo_coded = geo_coded.drop(self.block_group, axis = 1)            
        elif self.census_tract is not None:
            geo_coded['GEOID_BG'] = np.nan
            geo_coded = geo_coded.drop(self.census_tract, axis = 1)
            geo_coded = geo_coded.merge(data[self.census_tract], right_index = True, left_index = True, how = 'left')
            print(geo_coded[self.census_tract] == '')
            geo_coded['GEOID_CT'] = np.where((geo_coded[self.census_tract].isna()) | (geo_coded[self.census_tract].str.contains("None") | (geo_coded[self.census_tract] == ''))
                                             ,geo_coded['GEOID_CT']
                                             ,geo_coded[self.census_tract])
            geo_coded = geo_coded.drop(self.census_tract, axis = 1)
                                                   
        print("")
        
        print("[Completed] Preparing geo data")
        print("")
        print("[Start] Preparing ACS data")
        
        print("   [Start] Validating ACS input data")
        validate = ValidateGeocoded()
        validate.fit()
        acs_validator = validate.transform(geo_coded)
        save_json(acs_validator, self.out_path, "input_acs_validator.json")
        print("   [Completed] Validating ACS input data")
        print("")
        amp = ACSModelPrep(**self.params_dict)
        amp.fit()
        data_out = amp.transform(geo_coded, False)
        print("[Complete] Preparing ACS data")
        print("")
        
        return(data_out)



from zrp.modeling.predict import BISGWrapper, ZRP_Predict
from os.path import dirname, join, expanduser
from zrp.prepare.prepare import ZRP_Prepare
from zrp.prepare.base import BaseZRP
from zrp.prepare.utils import *
import pandas as pd
import numpy as np
import warnings
import surgeo
import pickle
import joblib
import json
import pycm
import sys
import os
import re


class ZRP(BaseZRP):
    """
    Zest Race Predictor, predicts race & ethnicity using name & geograhpic data
    
    Parameters
    ----------

    support_files_path: "data/processed"
        File path with support data.
    key: str, default 'ZEST_KEY', 
        Key to set as index. Optional. If not provided, a key will be generated.
    first_name: str, default 'first_name'
        Name of first name column
    middle_name: str, default 'middle_name'
        Name of middle name column
    last_name: str, default 'last_name'
        Name of last name/surname column
    house_number: str, default 'house_number'
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line. Optional to not provide if Census GEOIDs are provided and no geocoding is required.
    street_address: str, default 'street_address'
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str, default 'city'
        Name of city column.
    state: str, default 'state'
        Name of state column.
    zip_code: str, default 'zip_code'
        Name of zip or postal code column.
    race: str, default 'race', optional
        Name of the race column
    census_tract: str, optional
        Name of Census tract column. Recommended to provide if available or a geocoder is typically implemented.
    block_group: str, optional
        Name of Census block group column. Recommended to provide if available or a geocoder is typically implemented.
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list, optional
        List of missing values to replace.
    file_path: str, optional
        Path where to put artifacts and other files generated during intermediate steps. 
    geocode: bool, default True
        Geocoding indicator, to be deprecated by version 0.4.0.
    bisg: bool, default True
        Whether to return BISG proxies. 
    readout: bool, default True
        Whether to return a readout.
    n_jobs: int, default 49
        Number of jobs in parallel
    year: str, default '2019'
        ACS year to use.
    span: str, default '5'
        Year span of ACS data to use.
    runname: str, default 'test'
    """

    def __init__(self, file_path=None, pipe_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.pipe_path = pipe_path
        self.params_dict =  kwargs

    def fit(self):
        return self
    
    def rename_data_columns(self, data):
        """
        Renames the user specified columns of the input data to the default column names expected by ZRP.

        Parameters
        -----------
        data: pd.Dataframe
            Dataframe to be transformed
        """
        renamed_columns = {self.first_name: "first_name", self.middle_name: "middle_name", self.last_name: "last_name", self.house_number: "house_number", self.street_address: "street_address", self.city: "city", self.state: "state", self.zip_code: "zip_code"}
        data = data.rename(columns=renamed_columns)
        # self.params_dict = {}
        return data
    
    def check_for_old_files(self):
        """
        Checks if there are no files created in previous runs.

        Parameters
        -----------
        """
        old_files = []
        if self.runname is not None:
            file_like_geo = f'Zest_Geocoded_{self.runname}__{self.year}__'
            file_like_zrp_proxy = f'proxy_output_{self.runname}.feather'
            file_like_bisg_proxy = f'bisg_proxy_output__{self.runname}.feather'
        else:
            file_like_geo = f'Zest_Geocoded__{self.year}__'
            file_like_zrp_proxy = 'proxy_output.feather'
            file_like_bisg_proxy = 'bisg_proxy_output.feather'
        
        for file in os.listdir(self.out_path):
            if file_like_geo in file:
                old_files.append(os.path.join(self.out_path,file))
           
        file = os.path.join(self.out_path,file_like_zrp_proxy)
        if os.path.exists(file):
            old_files.append(file)
        
        file = os.path.join(self.out_path,file_like_bisg_proxy)
        if os.path.exists(file):
            old_files.append(file)
        
        if len(old_files) > 0:
            raise Exception(f"New value of 'runname' parameter needs to be specified or the following files need to be moved or deleted: {old_files}")
        
                  
    
    def transform(self, input_data, chunk_size = 25000):
        """
        Processes input data and generates ZRP predictions. Generates BISG predictions additionally if specified.

        Parameters
        -----------
        input_data: pd.Dataframe
            Dataframe to be transformed
        chunk_size: int
            Numer of rows to be processed in each iteration. input_data processed all at once if None is provided.
        """
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.file_path)

        data = self.rename_data_columns(data)
        self.reset_column_names()
        
        make_directory(self.out_path)
        self.check_for_old_files()
        curpath = dirname(__file__)
        if self.pipe_path is None:
            self.pipe_path = join(curpath, "modeling/models")

        data = data.sort_values('state')
        if chunk_size is None:
            chunk_size = len(data)
        chunk_max = int((len(data)-1)/chunk_size) + 1
        predict_out_list = list()
        full_bisg_proxies_list = list()
        for chunk in range(chunk_max):
            print("####################################")
            print(f'Processing rows: {chunk*chunk_size}:{(chunk+1)*chunk_size}')
            print("####################################")
            data_chunk = data[chunk*chunk_size:(chunk+1)*chunk_size]
            z_prepare = ZRP_Prepare(file_path=self.file_path, **self.params_dict)
            z_prepare.fit(data_chunk)
            prepared_data_chunk = z_prepare.transform(data_chunk)

            z_predict = ZRP_Predict(file_path=self.file_path, pipe_path=self.pipe_path, **self.params_dict)
            z_predict.fit(prepared_data_chunk)
            predict_out_chunk = z_predict.transform(prepared_data_chunk, save_table = False)
            predict_out_list.append(predict_out_chunk)
            
            if self.bisg:
                bisgw = BISGWrapper(**self.params_dict)
                bisg_proxies_chunk = bisgw.transform(prepared_data_chunk[~prepared_data_chunk.index.duplicated(keep='first')])
                full_bisg_proxies_list.append(bisg_proxies_chunk)
           
        predict_out = pd.concat(predict_out_list)
        
        if self.runname is not None:
            file_name = f'proxy_output_{self.runname}.feather'
        else:
            file_name = 'proxy_output.feather'
        save_feather(predict_out, self.out_path, file_name)    
        if self.bisg:
            full_bisg_proxies = pd.concat(full_bisg_proxies_list)
            if self.runname is not None:
                file_name = f'bisg_proxy_output_{self.runname}.feather'
            else:
                file_name = 'bisg_proxy_output.feather'
            save_feather(full_bisg_proxies, self.out_path, file_name) 

        try:
            predict_out = input_data.merge(predict_out.reset_index(drop=False), on=self.key)
        except KeyError:
            pass

        return (predict_out)


from os.path import dirname, join, expanduser
from zrp.prepare.preprocessing import *
from zrp.prepare.geo_geocoder import *
from zrp.prepare.acs_mapper import *
from zrp.prepare.base import BaseZRP
from zrp.prepare.utils import *
from zrp.validate import *
import pandas as pd
import numpy as np
import warnings
import glob
import json
import tqdm
import sys
import os
import re
import pickle
import joblib
import surgeo
import xgboost

import warnings
warnings.filterwarnings(action='ignore')

from zrp.modeling import src
from zrp.modeling.src.app_preprocessor import HandleCompoundNames
from zrp.modeling.src.acs_scaler import CustomRatios
from zrp.modeling.src.app_fe import AppFeatureEngineering, NameAggregation


class PredictPass(BaseZRP):
    """
    Generates proxies

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    """

    def __init__(self, pipe_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        
    def fit(self):
        return self
    
    def transform(self, input_data):
        """
        Processes input data and generates BISG predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
        proxies = pd.DataFrame({"AAPI": None, "AIAN": None, "BLACK": None,
                                "HISPANIC": None, "WHITE": None, f"{self.race}_proxy": None}, index = data.index)
        return(proxies) 

def validate_case(data, key, last_name):
    df = data.copy()
    new_row = df.tail(1).reset_index(drop=False)
    new_row[key] = "validate_case_001"
    new_row[last_name] = "SMITH JONES"
    new_row = new_row.set_index(key)
    df = pd.concat([df, new_row])
    return(df)
    
def validate_drop(data):
    data = data.drop('validate_case_001')
    return(data) 


class BISGWrapper(BaseZRP):
    """
    Wrapper function for Bayesian Improved Surname Geocoding

    Generates proxies using BISG algorithm.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        
    def fit(self, input_data):
        """

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """        
        if (self.last_name not in input_data.columns):
            raise ValueError('Last name needs to be provided when initializing this class. Please provide last name data in a column named "last_name" or set the custom name of the last name column in the data')
        if self.zip_code not in input_data.columns:
            raise ValueError('Zip or postal code name needs to be provided when initializing this class. Please provide zip code data in a column named "zip_code" or set the custom name of the zip code column in the data')
        return self
    
    def transform(self, input_data):
        """
        Processes input data and generates BISG predictions.

        Parameters
        -----------
        data: pd.Dataframe
            Dataframe to be transformed
        """
        df = input_data.copy()
        df = df[~df.index.duplicated(keep='first')]

        df = df.filter([self.last_name, self.zip_code, self.census_tract])
        df[self.zip_code] = np.where((df[self.zip_code] == "") | (df[self.zip_code] == " ") | (df[self.zip_code] == None) | (df[self.zip_code].str.len()<5), "99999", df[self.zip_code])
        
        bisg = surgeo.SurgeoModel()
        bisg_results = bisg.get_probabilities(names  = df[self.last_name].reset_index(drop=True),
                                              geo_df = df[self.zip_code].astype(int))
        combo = df.reset_index().merge(bisg_results, 
                           how="left", 
                           left_on=[self.last_name, self.zip_code], 
                           right_on=["name", "zcta5"]
                          )
        combo.drop(columns = ['zcta5', 'name'], inplace=True)
        
        combo.rename(columns = {
            'white': 'WHITE',
            'black': 'BLACK',
            'api': 'AAPI',
            'native': 'AIAN',
            'multiple': 'OTHER',
            'hispanic': 'HISPANIC'
        }, inplace=True) 
        combo = combo.set_index(self.key)
        combo = combo[~combo.index.duplicated(keep='first')]
                
        # Generate proxy at threshold
        race_col_ov = ['WHITE', 'BLACK', 'AAPI', 'AIAN', 'HISPANIC']
        subset = combo.filter(race_col_ov)
        subset = subset[race_col_ov].div(subset[race_col_ov].sum(axis=1), axis=0)              
        identifiedRaces = subset.idxmax(axis=1)
        combo[f"{self.race}_proxy"] = identifiedRaces
        combo['source_bisg'] = 1
        proxies = combo.filter(["AAPI", "AIAN", "BLACK", "HISPANIC",
                         "WHITE", f"{self.race}_proxy", "source_bisg"])   
        return(proxies)



class ZRP_Predict_ZipCode(BaseZRP):
    """
    Generates proxies using model trained on zip code features.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    """

    def __init__(self, pipe_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        
    def fit(self):
        return self
    
    def transform(self, input_data, geo_only=False):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """

        src_path = os.path.join(self.pipe_path, "zip_code")
        sys.path.append(src_path)        
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
            
        numeric_cols = list(data.filter(regex='^B|^C16').columns)
        data[numeric_cols] = data[numeric_cols].apply(pd.to_numeric, errors='coerce') 
        
        model = xgboost.Booster()
        model.load_model(os.path.join(src_path,"model.txt"))
#         model = pd.read_pickle(os.path.join(src_path,"model.pkl") )
        pipe = pd.read_pickle(os.path.join(src_path, "pipe.pkl") )
        
        
        data = validate_case(data, self.key, self.last_name)
        fe_data = pipe.transform(data)
        fe_data = validate_drop(fe_data)
        fe_matrix = xgboost.DMatrix(fe_data)
        
        proxies = pd.DataFrame(model.predict(fe_matrix), index = fe_data.index)
#         proxies.columns = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        proxies.columns = sorted(pipe.steps[2][1].mlb_columns)
        proxies[f"{self.race}_proxy"] = proxies.idxmax(axis=1)
        if not geo_only:
            proxies['source_zrp_zip_code'] = 1
        else:
            proxies['source_zrp_zip_code_geo_only'] = 1
        return(proxies)
        


class ZRP_Predict_BlockGroup(BaseZRP):
    """
    Generates proxies using model trained on block group features.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    """

    def __init__(self, pipe_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        
    def fit(self):
        return self
    
    def transform(self, input_data, geo_only=False, name_only=False):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed.
        geo_only: Boolean
            Whether or not the predictions will be based on block group geo features only.
        name_only: Boolean
            Whether or not the predictions will be based on name features only.
        """        

        src_path = os.path.join(self.pipe_path,"block_group")
        sys.path.append(src_path)
        
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)

        numeric_cols = list(data.filter(regex='^B|^C16').columns)
        data[numeric_cols] = data[numeric_cols].apply(pd.to_numeric, errors='coerce')
        
        model = xgboost.Booster()
        model.load_model(os.path.join(src_path,"model.txt"))
#         model = pd.read_pickle(os.path.join(src_path,"model.pkl") )
        pipe = pd.read_pickle(os.path.join(src_path, "pipe.pkl") )
        data = validate_case(data, self.key, self.last_name)
        fe_data = pipe.transform(data)
        fe_data = validate_drop(fe_data)
        fe_matrix = xgboost.DMatrix(fe_data)
        proxies = pd.DataFrame(model.predict(fe_matrix), index = fe_data.index)
#         proxies.columns = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        proxies.columns = sorted(pipe.steps[2][1].mlb_columns)
        proxies[f"{self.race}_proxy"] = proxies.idxmax(axis=1)
        
        if not geo_only and not name_only:
            proxies['source_zrp_block_group'] = 1
        elif geo_only:
            proxies['source_zrp_block_group_geo_only'] = 1
        else:
            proxies['source_zrp_name_only'] = 1
        return(proxies)
        



class ZRP_Predict_CensusTract(BaseZRP):
    """
    Generates proxies using model trained on census tract features.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    """

    def __init__(self, pipe_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        
    def fit(self):
        return self
    
    def transform(self, input_data, geo_only=False):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """        

        src_path = os.path.join(self.pipe_path,"census_tract")
        sys.path.append(src_path)
        
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
            
        numeric_cols = list(data.filter(regex='^B|^C16').columns)
        data[numeric_cols] = data[numeric_cols].apply(pd.to_numeric, errors='coerce') 

        model = xgboost.Booster()
        model.load_model(os.path.join(src_path,"model.txt"))        
#         model = pd.read_pickle(os.path.join(src_path,"model.pkl") )
        pipe = pd.read_pickle(os.path.join(src_path, "pipe.pkl") )
        
        data = validate_case(data, self.key, self.last_name)
        fe_data = pipe.transform(data)
        fe_data = validate_drop(fe_data)
        fe_matrix = xgboost.DMatrix(fe_data)
        proxies = pd.DataFrame(model.predict(fe_matrix), index = fe_data.index)
#         proxies.columns = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        proxies.columns = sorted(pipe.steps[2][1].mlb_columns)
        proxies[f"{self.race}_proxy"] = proxies.idxmax(axis=1)
        if not geo_only:
            proxies['source_zrp_census_tract'] = 1
        else:
            proxies['source_zrp_census_tract_geo_only'] = 1
        return(proxies)
    


class ZRP_Predict(BaseZRP):
    """
    Generates race proxies.
    Attempts to predict on block group, then census tract, then zip code based on which level ACS data is found for. If
    Geo level data is unattainable, the BISG proxy is computed. No prediction returned if BISG cannot be computed either.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    file_path: str
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    """

    def __init__(self, pipe_path, file_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.pipe_path = pipe_path
        self.params_dict = kwargs
        self.census_tract = 'GEOID_CT'
        self.block_group = 'GEOID_BG'
        self.zip_code = 'GEOID_ZIP'
        
        
    def fit(self, data):
        if xgboost.__version__ != "1.0.2":
            raise AssertionError("XGBoost version does not match requirements, required version is 1.0.2")

        data_cols =  list(data.columns)
        self.required_cols = [self.first_name, self.middle_name, self.last_name, "GEOID", "B01003_001"]
        val_na = is_missing(data, self.required_cols)
        if val_na:
            raise ValueError(f"Missing required data {val_na}")     

        print("   [Start] Validating pipeline input data")
        validator = ValidateInput()
        validator.fit()
        validator_in = validator.transform(data)
        save_json(validator_in, self.out_path, "input_predict_validator.json")
        print("   [Completed] Validating pipeline input data")
        print("")
        return self
    
    def validate_data_has_names(self, data, is_input = True):
        """Identifies which rows have missing "first_name", "middle_name", or "last_name". Returns a df with columns for the 3 names and rows taking boolean
        values for whether or not this record contains said name or not.
        
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        is_input: bool
            Indicator if validating raw input data
        """
        name_cols =  [self.first_name, self.middle_name, self.last_name]
        
        data = data.copy()
        data["has_first_name"] = 0
        data["has_middle_name"] = 0
        data["has_last_name"] = 0
        has_name_columns = ["has_first_name", "has_middle_name", "has_last_name"]
        
        for name_col, has_name_col in zip(name_cols, has_name_columns):
            data.loc[~((data[name_col].astype(str).str.upper() == "NONE")
                                | (data[name_col].astype(str).str.upper() == " ")
                                | (data[name_col].isna()))
                , has_name_col] = 1 
#         return(data[has_name_columns])
        return(data)

    def standard_target_classes(self):
        """
        Checks if models were trained on standard set of target classes. Returns True for standard target classes and False for non-standard target classes. 
        
        Parameter
        ---------
        """
        model_types = ['block_group', 'census_tract', 'zip_code']
        all_model_cols = set()
        for model_type in model_types:
            src_path = os.path.join(self.pipe_path, model_type)
            sys.path.append(src_path)
            pipe = pd.read_pickle(os.path.join(src_path, "pipe.pkl"))
            cols = set(pipe.steps[2][1].mlb_columns)
            all_model_cols = all_model_cols.union(cols)
        standard_cols = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        target_is_standard = set(all_model_cols) == set(standard_cols)
        return(target_is_standard)

    def transform(self, input_data, save_table=True):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """        

        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
        cur_path = dirname(__file__)

        flb = load_json(f'{cur_path}/feature_list_bg.json')
        flc = load_json(f'{cur_path}/feature_list_ct.json')
        flz = load_json(f'{cur_path}/feature_list_zp.json')
        
        # Determine records with names provided
        data = self.validate_data_has_names(input_data)
                    
        if 'acs_source' in data.columns:  
            # Select records for zrp geo and name proxy
            df_0 = data[
                (data['acs_source'] == 'BG') &
                ((data["has_first_name"] == 1) |
                 (data["has_middle_name"] == 1) |
                 (data["has_last_name"] == 1)
                
            )]
            keys_0 = list(df_0.index.values)
            
            df_1 = data[
                ~(data.index.isin(keys_0)) &
                (data['acs_source'] == 'CT') &
                ((data["has_first_name"] == 1) |
                 (data["has_middle_name"] == 1) |
                 (data["has_last_name"] == 1)
                
            )]
            keys_1 = keys_0 + list(df_1.index.values)
            
            df_2 = data[
                ~(data.index.isin(keys_1)) &
                (data['acs_source'] == 'ZIP') &
                ((data["has_first_name"] == 1) |
                 (data["has_middle_name"] == 1) |
                 (data["has_last_name"] == 1)
                
            )]
            keys_2 = keys_1 + list(df_2.index.values)
            
            # Select records for zrp geo only proxies
            df_3 = data[
                ~(data.index.isin(keys_2)) &
                (data['acs_source'] == 'BG') &
                ~((data["has_first_name"] == 1) &
                 (data["has_middle_name"] == 1) &
                 (data["has_last_name"] == 1)
            )]
            keys_3 = keys_2 + list(df_3.index.values)
            
            df_4 = data[
                ~(data.index.isin(keys_3)) &
                (data['acs_source'] == 'CT') &
                ~((data["has_first_name"] == 1) &
                 (data["has_middle_name"] == 1) &
                 (data["has_last_name"] == 1)
                
            )]
            keys_4 = keys_3 + list(df_4.index.values)
            
            df_5 = data[
                ~(data.index.isin(keys_4)) &
                (data['acs_source'] == 'ZIP') &
                ~((data["has_first_name"] == 1) &
                 (data["has_middle_name"] == 1) &
                 (data["has_last_name"] == 1)
                
            )]
            keys_5 = keys_4 + list(df_5.index.values)
            
            # Records for BISG, zrp name only, or No-proxy
            df_6 = data[~(data.index.isin(keys_5))]
            keys_6 = keys_5 + list(df_6.index.values)
            
            # Predict BISG
            # Check the indices where proxy is none
            # run ZRP name only on these left over
            # check where proxy is none
            # mark that these have source no_proxy
            
        else:
            raise KeyError("Processed data is required for ZRP_Predict. Please use EnginetoPredict if supplying feature engineered data to generate predictions" )
            
        out_list = []
        if not df_0.empty:    # BG & Names
            zrp_bg_complete = ZRP_Predict_BlockGroup(self.pipe_path, **self.params_dict)
            out_0 = zrp_bg_complete.transform(df_0.filter(flb))
            out_list.append(out_0)    
        if not df_1.empty:    # CT & Names
            zrp_ct_complete = ZRP_Predict_CensusTract(self.pipe_path, **self.params_dict)
            out_1 = zrp_ct_complete.transform(df_1.filter(flc))
            out_list.append(out_1)    
        if not df_2.empty:    # ZC & Names
            zrp_zp_complete = ZRP_Predict_ZipCode(self.pipe_path, **self.params_dict)
            out_2 = zrp_zp_complete.transform(df_2.filter(flz))
            out_list.append(out_2)    
        if not df_3.empty:    # BC Only
            zrp_bg_geo_only = ZRP_Predict_BlockGroup(self.pipe_path, **self.params_dict)
            out_3 = zrp_bg_geo_only.transform(df_3.filter(flb), geo_only=True)
            out_list.append(out_3)  
        if not df_4.empty:    # CT Only
            zrp_ct_geo_only = ZRP_Predict_CensusTract(self.pipe_path, **self.params_dict)
            out_4 = zrp_ct_geo_only.transform(df_4.filter(flc), geo_only=True)
            out_list.append(out_4)   
        if not df_5.empty:    # ZC Only
            zrp_zp_geo_only = ZRP_Predict_ZipCode(self.pipe_path, **self.params_dict)
            out_5 = zrp_zp_geo_only.transform(df_5.filter(flz), geo_only=True)
            out_list.append(out_5)  
        if not df_6.empty:    # BISG
            bisgw = BISGWrapper(**self.params_dict)
            bisgw.fit(df_6)
            out_6 = bisgw.transform(df_6)

            if self.standard_target_classes():
                # Capture successful BISG proxies 
                bisg_proxies = out_6[~(out_6[f"{self.race}_proxy"].isna())]
                out_list.append(bisg_proxies) 
                
                # Records that failed BISG proxy
                records_failed_bisg_proxy = df_6.loc[df_6.index.intersection(out_6[out_6[f"{self.race}_proxy"].isna()].index.values)]
            else:
                # For non-standard classes disregard BISG predictions 
                records_failed_bisg_proxy = df_6
            
            if not records_failed_bisg_proxy.empty:    # Attempt name only ZRP proxying
                df_7 = records_failed_bisg_proxy[
                    (records_failed_bisg_proxy['has_first_name'] == 1) |
                    (records_failed_bisg_proxy['has_middle_name'] == 1) |
                    (records_failed_bisg_proxy['has_last_name'] == 1)
                ]     
                if not df_7.empty:
                    zrp_names_only = ZRP_Predict_BlockGroup(self.pipe_path, **self.params_dict)
                    out_7 = zrp_names_only.transform(df_7.filter(flb), name_only=True)
                    out_list.append(out_7)  

                    record_indices_for_name_only_proxy = list(df_7.index.values)
                    cannot_proxy_records = records_failed_bisg_proxy[~(records_failed_bisg_proxy.index.isin(record_indices_for_name_only_proxy))]
                else:
                    cannot_proxy_records = records_failed_bisg_proxy

                if not cannot_proxy_records.empty:
                    failed_proxies = out_6.loc[out_6.index.intersection(cannot_proxy_records.index.values)]
                    failed_proxies.rename(columns={"source_bisg": "source_no_proxy"}, inplace=True)
                    out_list.append(failed_proxies)
                
        if df_6.empty or records_failed_bisg_proxy.empty or cannot_proxy_records.empty:
            print("   ...Proxies generated")

        proxies_out = pd.concat(out_list)
        
        # Rearegement of columns    
        all_source_cols = [
            'source_zrp_block_group', 'source_zrp_census_tract',
            'source_zrp_zip_code', 'source_bisg', 'source_zrp_block_group_geo_only',
            'source_zrp_census_tract_geo_only', 'source_zrp_zip_code_geo_only',
            'source_zrp_name_only', 'source_no_proxy']
        source_cols = [col for col in all_source_cols if col in proxies_out.columns]
        race_cols = list(set(proxies_out.columns) - set(source_cols) - set([f"{self.race}_proxy"]))
        race_cols.sort()
        ordered_columns = race_cols + [f"{self.race}_proxy"] + source_cols  
        proxies_out = proxies_out[ordered_columns]
        
        proxies_out[race_cols] = proxies_out[race_cols].fillna(0)
        proxies_out[source_cols] = proxies_out[source_cols].fillna(0)
        proxies_out = proxies_out.sort_values(source_cols)        
        
        if save_table:
            make_directory(self.out_path)
            if self.runname is not None:
                file_name = f'proxy_output_{self.runname}.feather'
            else:
                file_name = 'proxy_output.feather'
            save_feather(proxies_out, self.out_path, file_name)      
        
        return(proxies_out)

    
    
class FEtoPredict(BaseZRP):
    """
    Generates proxies from feature engineered data.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    pipe_type: str, (default='census_tract')
        Type of pipeline that generated the engineered data.
        Options: 'block_group', 'census_tract', or 'zip_code'
    """

    def __init__(self, pipe_path, pipe_type='census_tract', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        self.pipe_type = pipe_type
        
        
    def fit(self):
        return self
    
    def transform(self, input_data, save_table=True):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """
        src_path = os.path.join(self.pipe_path, self.pipe_type)
        model = xgboost.Booster()
        model.load_model(os.path.join(src_path,"model.txt"))
#         model = pd.read_pickle(os.path.join(self.pipe_path, f"{pipe_type}/model.pkl") )
        pipe = pd.read_pickle(os.path.join(src_path, "pipe.pkl"))
        # Load Data
        if input_data.empty:
            raise ValueError("Feature engineered data is empty or missing. Please provide the feature engineered data as `input_data` to generate predictions.")
        fe_data = input_data.copy()
        fe_data = fe_data.set_index('ZEST_KEY')
        fe_matrix = xgboost.DMatrix(fe_data)

        proxies = pd.DataFrame(model.predict(fe_matrix), index = fe_data.index)
        proxies.columns = sorted(pipe.steps[2][1].mlb_columns)
        proxies[f"{self.race}_proxy"] = proxies.idxmax(axis=1)
        proxies[f'source_{self.pipe_type}'] = 1        

        
        if save_table:
            make_directory(self.out_path)
            if self.runname is not None:
                file_name = f'proxy_output_{self.runname}.feather'
            else:
                file_name = 'proxy_output.feather'
            save_feather(proxies_out, self.out_path, file_name) 
            
        return(proxies)
    



from zrp.prepare.preprocessing import *
from os.path import dirname, join, expanduser
from zrp.prepare.utils import *
from joblib import Parallel, delayed
import pandas as pd
import numpy as np
import fiona
import json
import sys
import os
import re

def gdbToDf_short(file, indx):
    """
    Loads shapefiles
    
    Parameters
    ----------
    file: str
        Filepath or name of file to load
    indx: str
        Indicates type of shapefile to load in
    """      
    dictList = []
    if indx == "address":
        with fiona.open(file) as src:
            for i in range(10000000):
                dictList.append(next(src)["properties"])
    else:
        with fiona.open(file) as src:
            for i in range(len(src)):
                dictList.append(next(src)["properties"])
    df = pd.DataFrame(
        dictList,
        columns=dictList[0].keys(),
    )
    return df

def sides_split(addr_edge_df):
    """Split tigerline dataframes into sides
    Parameters
    ----------
    addr_edge_df: pd.DataFrame/tigerline dataframe with address features & edges data
    """
    l_addr_edge_df = addr_edge_df.copy()
    r_addr_edge_df = addr_edge_df.copy()

    l_addr_edge_df = l_addr_edge_df.drop(['TFIDR', 'ARIDR', 'RFROMHN',
                                           'RTOHN', 'ZIPR', 'PARITYR',
                                           'PLUS4R', 'RFROMTYP', 'RTOTYP',
                                           'OFFSETR', 'RFROMADD', 'RTOADD'],
                                         axis=1)
    r_addr_edge_df = r_addr_edge_df.drop(['TFIDL', 'ARIDL', 'LFROMHN', 
                                          'LTOHN', 'ZIPL', 'PARITYL',
                                          'PLUS4L', 'LFROMTYP', 'LTOTYP',
                                          'OFFSETL', 'LFROMADD', 'LTOADD'],
                                         axis=1)

    l_addr_edge_df["SIDE"] = "L"
    r_addr_edge_df["SIDE"] = "R"

    l_addr_edge_df = l_addr_edge_df.rename(
        columns={"TFIDL": "TFID",
                 "ARIDL": "ARID",
                 "LFROMHN": "FROMHN",
                 "LTOHN": "TOHN",
                 "ZIPL": "ZIP",
                 "PARITYL": "PARITY",
                 "PLUS4L": "PLUS4",
                 "LFROMTYP":"FROMTYP",
                 "LTOTYP": "TOTYP",
                 "OFFSETL": "OFFSET",
                 "LFROMADD": "FROMADD",
                 "LTOADD": "TOADD" }
    )
    r_addr_edge_df = r_addr_edge_df.rename(
        columns={"TFIDR": "TFID",
                 "ARIDR": "ARID",
                 "RFROMHN": "FROMHN",
                 "RTOHN": "TOHN",
                 "ZIPR": "ZIP",
                 "PARITYR": "PARITY",
                 "PLUS4R": "PLUS4",
                 "RFROMTYP":"FROMTYP",
                 "RTOTYP": "TOTYP",
                 "OFFSETR": "OFFSET",
                 "RFROMADD": "FROMADD",
                 "RTOADD": "TOADD"}
    )
    return (l_addr_edge_df, r_addr_edge_df)


def concat_shapes(df_list):
    """Concatenate lists of shapefile data"""
    df = pd.concat(df_list, axis=0)
    return (df)


def merge_shapes(df_1, df_2, merge_cols, merge_method):
    """Merge shapefile dataframes"""
    df = df_1.merge(df_2, on=merge_cols, how=merge_method)
    return (df)


class GeoLookUpBuilder():
    """
    This class builds geo lookup tables.
    
    Parameters
    ----------
    support_files_path: str
        Path to support files
    year: str (default 2019)
        Year associated with ACS data
    output_folder_suffix: str
        Suffix attached to the output folder name.
    """

    def __init__(self, support_files_path, year, output_folder_suffix):
        self.support_files_path = support_files_path
        self.year = year
        self.output_folder_suffix = output_folder_suffix
        self.raw_geo_path = os.path.join(self.support_files_path, "raw/geo", self.year)
        self.out_geo_path = os.path.join(self.support_files_path, "processed/geo", f"{self.year}_{self.output_folder_suffix}")

    def fit(self):
        return self

    def transform(self, st_cty_code, save_table=True):
        """
        Returns DataFrame of geo lookup tables

        Parameters
        ----------
        st_cty_code: str
            state county code string
        save_table: bool
            Indicator to save dataframe
        """
        curpath = dirname(__file__)
        make_directory(output_directory=self.raw_geo_path)
        make_directory(output_directory=self.out_geo_path)

        geo_support_files_path = self.raw_geo_path
        print("Shapefile input:", geo_support_files_path)
        print("Lookup Table output:", self.out_geo_path)
        print("")

        addrfeat = "".join([geo_support_files_path,
                            "/tl_",
                            self.year,
                            "_",
                            st_cty_code,
                            "_addrfeat.shp"])
        faces = "".join([geo_support_files_path,
                         "/tl_",
                         self.year,
                         "_",
                         st_cty_code,
                         "_faces.shp"])
        edges = "".join([geo_support_files_path,
                         "/tl_",
                         self.year,
                         "_",
                         st_cty_code,
                         "_edges.shp"])

        print(" ... Loading requirements")
        af_df = gdbToDf_short(addrfeat,
                              None)
        fc_df = gdbToDf_short(faces,
                              "face")
        ed_df = gdbToDf_short(edges,
                              "edge")

        af_col_keep = [
            "TLID",
            "TFIDL",
            "TFIDR",
            "ARIDL",
            "ARIDR",
            "LINEARID",
            "FULLNAME",
            "LFROMHN",
            "LTOHN",
            "RFROMHN",
            "RTOHN",
            "ZIPL",
            "ZIPR",
            "EDGE_MTFCC",
            "ROAD_MTFCC",
            "PARITYL",
            "PARITYR",
            "LFROMTYP",
            "LTOTYP",
            "RFROMTYP",
            "RTOTYP",
            "OFFSETL",
            "OFFSETR",
            "PLUS4R",
            "PLUS4L"
        ]

        if "STATEFP20" in fc_df.columns:
            fc_df = fc_df.rename(columns={"STATEFP20": "STATEFP",
                                          "COUNTYFP20": "COUNTYFP",
                                          "TRACTCE20": "TRACTCE",
                                          "BLKGRPCE20": "BLKGRPCE",
                                          "BLOCKCE20": "BLOCKCE",
                                          "TTRACTCE20": "TTRACTCE",
                                          "TBLKGPCE20": "TBLKGPCE",
                                          "ZCTA5CE20": "ZCTA5CE",
                                          "PUMA5CE20": "PUMACE",
                                          "PUMA5CE10": "PUMACE10"
                                          }
                                 )
        else:
            fc_df[["STATEFP", "COUNTYFP", "TRACTCE", "BLKGRPCE", "BLOCKCE", "ZCTA5CE", "PUMACE"]] = fc_df[
                ["STATEFP10", "COUNTYFP10", "TRACTCE10", "BLKGRPCE10", "BLOCKCE10", "ZCTA5CE10", "PUMACE10"]]

        fc_col_keep = [
            "TFID",
            "STATEFP10",
            "COUNTYFP10",
            "TRACTCE10",
            "BLKGRPCE10",
            "BLOCKCE10",
            "ZCTA5CE10",
            "PUMACE10",
            "STATEFP",
            "COUNTYFP",
            "TRACTCE",
            "BLKGRPCE",
            "BLOCKCE",
            "ZCTA5CE",
            "TTRACTCE",
            "TBLKGPCE",
            "OFFSET",
            "PUMACE"
        ]

        ed_col_keep = [
            "STATEFP",
            "COUNTYFP",
            "TLID",
            "TFIDL",
            "TFIDR",
            "FULLNAME",
            "LFROMADD",
            "LTOADD",
            "RFROMADD",
            "RTOADD",
            "ZIPL",
            "ZIPR",
            "OFFSETL",
            "OFFSETR",
        ]

        af_df = af_df[af_col_keep]
        fc_df = fc_df[fc_col_keep]
        ed_df = ed_df[ed_col_keep]

        print(" ... Creating lookup table")

        addr_edge_df = merge_shapes(af_df,
                                    ed_df,
                                    ["TLID", "TFIDL", "TFIDR",
                                     "ZIPL", "ZIPR", "FULLNAME",
                                     "OFFSETL", "OFFSETR"],
                                    "inner")
        
        l_addr_edge_df, r_addr_edge_df = sides_split(addr_edge_df)

        l_addr_edge_face_df = merge_shapes(l_addr_edge_df,
                                           fc_df,
                                           ["TFID", "COUNTYFP", "STATEFP", "OFFSET"],
                                           "inner")
        r_addr_edge_face_df = merge_shapes(r_addr_edge_df,
                                           fc_df,
                                           ["TFID", "COUNTYFP", "STATEFP", "OFFSET"],
                                           "inner")
        
        l_addr_edge_face_df = l_addr_edge_face_df[l_addr_edge_face_df["ZIP"].notna()] 
        r_addr_edge_face_df = r_addr_edge_face_df[r_addr_edge_face_df["ZIP"].notna()]         

        aef = concat_shapes([l_addr_edge_face_df,
                             r_addr_edge_face_df])

        aef.drop_duplicates(inplace=True)

        print(" ... Formatting lookup table")
        # rename features to avoid user overlap
        aef = aef.rename(columns={"ZIP": "ZEST_ZIP",
                                  "FULLNAME": "ZEST_FULLNAME"})

        aef["RAW_ZEST_ZIP"] = aef["ZEST_ZIP"].copy()
        aef["RAW_ZEST_STATEFP"] = aef["STATEFP"].copy()
        aef["RAW_ZEST_COUNTYFP"] = aef["COUNTYFP"].copy()
        aef["RAW_ZEST_FULLNAME"] = aef["ZEST_FULLNAME"].copy()
        aef["RAW_ZEST_TRACTCE"] = aef["TRACTCE"].copy()
        aef["RAW_ZEST_BLKGRPCE"] = aef["BLKGRPCE"].copy()

        aef = aef.astype(str)
        gps = ProcessGLookUp()

        data_path = join(curpath, '../data/processed/state_mapping.json')
        state_mapping = load_json(data_path)

        aef = gps.transform(aef, state_mapping)    
        aef = aef.sort_values(['PARITY', 'ZEST_ZIP', 'ZEST_FULLNAME', 'FROMHN_LEFT', 'FROMHN_RIGHT'])                    
        aef = aef.reset_index(drop = True)
        
        final_cols = ['STATEFP', 'COUNTYFP', 'TRACTCE', 'BLKGRPCE', 'ZEST_FULLNAME', 'FROMHN' , 'TOHN', 'ZEST_ZIP', 
                      'ZCTA5CE', 'ZCTA5CE10', 'FROMHN_LEFT', 'FROMHN_RIGHT', 'TOHN_LEFT', 'TOHN_RIGHT', 'PARITY']
        aef = aef[final_cols]                    
        # optional save 
        if save_table:
            make_directory(self.out_geo_path)
            filename = "".join(["Zest_Geo_Lookup_", self.year, "_", st_cty_code, ".parquet"])
            print(f"Saving {filename}")
            save_dataframe(aef, self.out_geo_path, filename)
        else:
            print("No tables were saved")

        return (aef)

class GeoLookUpLooper(GeoLookUpBuilder):
    """
    This class loops through GeoLookUpBuilder() to build geo lookup tables
    
    Parameters
    ----------
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def fit(self):
        """
        Collects a dictionary of files to loop through

        Parameters
        ----------
        """          
        self.st_dict = dict()
        shp_list = [shp_file for shp_file in os.listdir(self.raw_geo_path) if ('_edges.shp' in shp_file
                                                                                or '_faces.shp' in shp_file
                                                                                or '_addrfeat.shp' in shp_file
                                                                                    )
                                                                                and '.iso.xml' not in shp_file]
        strings_count = pd.Series(shp_list).str[8:13].value_counts()
        st_cty_code_list = list(strings_count[strings_count == 3].sort_index().index)

        st_code_list = [st[0:2] for st in st_cty_code_list]
        st_code_list = np.unique(np.array(st_code_list))

        for st_code in st_code_list:
            self.st_dict[st_code] = [st_cty for st_cty in st_cty_code_list if st_cty[:2] == st_code]
        return self

    def __st_cty_builder(self, st_cty_code, geo_build, save_table):
        """
        Builds one lookup table on st_cty_code level

        Parameters
        ----------
        """       
        try:
            geo_build.transform(st_cty_code, save_table = save_table)
        except: 
            pass
    
    def transform(self, st_list = None, save_st_cty_tables=True, save_st_tables=True, n_jobs = -1):
        """
        Returns DataFrame of geo lookup tables

        Parameters
        ----------
        st_list: list
            list of state codes to process
        save_st_cty_tables: bool
            Indicator to save state-county level tables
        save_st_tables: bool
            Indicator to save state level tables
        n_jobs: int (default -1)
            Number of jobs in parallel
        """        

        if st_list is None:
            temp_st_dict = self.st_dict
        else:
            temp_st_dict = {key: self.st_dict[key] for key in st_list}
        
        if save_st_cty_tables:
            geo_build = GeoLookUpBuilder(support_files_path = self.support_files_path, 
                                     year = self.year, 
                                     output_folder_suffix = self.output_folder_suffix)
            for st_code in temp_st_dict:
                print(f"Working on st_cty tables (st_code = {st_code})")
                Parallel(n_jobs = n_jobs, verbose=1)(delayed(self.__st_cty_builder)(st_cty_code, geo_build, save_st_cty_tables) for st_cty_code in temp_st_dict[st_code])

        if save_st_tables:
            for st_code in temp_st_dict:
                print(f"Working on st tables (st_code = {st_code})")
                output_list = []
                for st_cty_code in temp_st_dict[st_code]:
                    try:
                        input_filename = "".join(["Zest_Geo_Lookup_", self.year, "_", st_cty_code, ".parquet"])
                        input_filepath = os.path.join(self.out_geo_path, input_filename)
                        output_list.append(pd.read_parquet(input_filepath))
                    except:
                        pass
                output = pd.concat(output_list).reset_index(drop = True)
                output_filename = "".join(["Zest_Geo_Lookup_", self.year, "_State_", st_code, ".parquet"])
                save_dataframe(output, self.out_geo_path, output_filename)

        return (self)
